// BUMP, the imageboard porter
// ===========================
// -This software is designed to help out anon with imageboards

// -Filename: Bump.cpp
// -The definitions of the Bump class, and it's ancillary Page class
// -Additional simple utility functions included

#include "Bump.h"

#include <algorithm>
#include <clocale>
#include <cstdio>  // temp
#include <cstdlib>
#include <cwctype>
#include <fstream>
#include <iomanip>
#include <ios>
#include <iostream>
#include <iterator>
#include <map>
#include <regex>
#include <sstream>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

#include "muh_curl.h"

using std::begin;
using std::cerr;
using std::end;
using std::fstream;
using std::get;
using std::ifstream;
using std::ios;
using std::istringstream;
using std::make_tuple;
using std::map;
using std::ofstream;
using std::pair;
using std::printf;  // temp
using std::string;
using std::to_string;
using std::tuple;
using std::vector;

// TODO:
// -Eventually, also include an MD5 general file & MD5 hashing to deal w/
//  the 'dup names, but not dup files' issue
// -Add thumbnail storage as well (during 'rewrite html files' phase)

//------------------------------------------------------------------------------

// Take a raw json source and parse it into a json db
template<typename T> Json::Value parse_json(T& json_src)
{
  Json::Reader reader{};
  Json::Value json_dat{};
  reader.parse(json_src, json_dat);
  return json_dat;
}

//------------------------------------------------------------------------------

// Load and parse a json file
Json::Value rd_json_file(const fs::path& filepath)
{
  ifstream json_file{filepath.string()};
  return parse_json(json_file);
}

//------------------------------------------------------------------------------

// Write json out to a file using no whitespace (fastwrite)
void fast_wrt_strm(std::fstream& io_strm, const Json::Value& json)
{
  // Prepare the stream for writing
  io_strm.clear();
  io_strm.seekp(0);

  Json::FastWriter writer;
  io_strm << writer.write(json);
}

//------------------------------------------------------------------------------

// Separate a string into two parts, dividing on the final char
// -currently '.'
std::tuple<std::string, std::string> split_parts(const std::string& input)
{
  // Search backwards to find the last '.'
  auto last_dot_pos{input.rfind('.')};

  if (last_dot_pos != string::npos) {
    const string front{input.substr(0, last_dot_pos)};
    const string back{input.substr(last_dot_pos + 1, string::npos)};
    return make_tuple(front, back);
  }
  else {  // dot not found
    return make_tuple(input, "");
  }
}

//------------------------------------------------------------------------------

// Consolidate consecutive char c to just one c
void rem_dups(std::string& in_out, const char c)
{
  char prev{'\0'};

  for (auto it = in_out.begin(); it != in_out.end(); ++it)
    if ((prev == *it) && (c == *it)) {
      in_out.erase(it);
      --it;
    }
    else
      prev = *it;
}

//------------------------------------------------------------------------------

// // Remove any leading or trailing char c
std::string trim_char(const std::string& in, const char c)
{
  string res{in};

  if (res[0] == c)  // Remove any leading c
    res = res.substr(1, res.size());

  if (res[res.size() - 1] == c)  // Remove any trailing c
    res = res.substr(0, (res.size() - 1));

  return res;
}

//------------------------------------------------------------------------------

std::string repl_w_char(const std::string& in,
                        const std::string& bad,
                        const char good)
{
  std::regex space_it{bad};
  std::string good_ch{good};

  return std::regex_replace(in, space_it, good_ch);
}

//------------------------------------------------------------------------------

std::string lower_cp(const std::string& input)
{
  string res{input};
  std::transform(
      begin(res), end(res), begin(res), [](unsigned char c) -> unsigned char {
        return std::tolower(c);
      });

  res = trim_char(res, ' ');

  return res;
}

//------------------------------------------------------------------------------

void alert_user(const std::string& page_uri, const std::string& err_msg = "")
{
  if (err_msg != "")
    fprintf(stderr,
            "\nBUMP got an online response data error: '%s'\n\n",
            err_msg.c_str());
  else
    fprintf(stderr, "\nBUMP got an online response data error\n\n");

  printf("Is  '%s'  supplying valid data? \n"
         " (this is the location BUMP is looking to for it's data)\n",
         page_uri.c_str());
}

//------------------------------------------------------------------------------

// DESIGN: Call this from exit_bad_status()
// -use tuple<string, uint> return here instead?

uint chk_resp_err(const std::string& raw_resp)
{
  uint err_code{0};
  string err_msg{};
  vector<pair<string, uint>> err_codes{{"403 Forbidden", 403},
                                       {"404 Not Found", 404},
                                       {"500 Internal Server Error", 500},
                                       {"502 Bad Gateway", 502}};

  // Search for an error code inside the response
  const auto& low_resp{lower_cp(raw_resp)};
  for (const auto& [msg, code] : err_codes) {
    auto resp_it{low_resp.find(lower_cp(msg))};
    if (resp_it != string::npos) {
      err_msg = msg;
      err_code = code;
      break;
    }
  }

  return err_code;
}

//------------------------------------------------------------------------------

Page::Page()
{
}

Page::Page(const std::string& site_name,
           const std::string& board_name,
           const std::string& page_name,
           const fs::path& target_dir,
           bool force_check,
           Json::Value board_json)
    : site_nm{site_name},
      board_nm{board_name},
      page_nm{page_name},
      targ_dir{target_dir},
      force_chk{force_check},
      brd_json{board_json}
{
  jsondb.parent = this;  // For error-reporting during load(), Const.

  // Basically, other than .onions, access should be via https
  auto [site, domain]{split_parts(site_nm)};
  if (domain != "onion")
    muh_curl::proto = "https";

  // DESIGN: Future adaptation for sites that don't provide a JSON API
  // if (site_has_json(site_nm))
  load_json();
  // else {
  //   load_html();
  //   const auto comp_json{compile_html_to_json()};
  //   load_json(comp_json);
  // }

  if (page_nm != "catalog")  // Skip 404 checking for the catalog Page
    load_404s();

  set_board_tags();
}

//------------------------------------------------------------------------------

// Get the live json and store it into the page's db
void Page::load_json()
{
  resp("json");
  jsondb.load(raw_json);
}

//------------------------------------------------------------------------------

// Store the board's live response inside the page
void Page::resp(const std::string& page_tp)
{
  page_uri = muh_curl::muh_url(site_nm, board_nm, page_nm, page_tp);
  printf(" page_uri:  <==  %s\n", page_uri.c_str());

  if (page_tp == "json")
    raw_json = muh_curl::get_resp(page_uri).str();
  else
    raw_html = muh_curl::get_resp(page_uri).str();
}

//------------------------------------------------------------------------------

// Parse the raw json from the site into the page's json db
void Page::Jsondb::load(const std::string& raw_json)
{
  Json::Reader reader{};

  if (!reader.parse(raw_json, db))
    rpt_bad_status(raw_json);
}

//------------------------------------------------------------------------------

// Search response for HTTP error code to report to user, then exit.
void Page::Jsondb::rpt_bad_status(const std::string& raw_resp) const
{
  // test:
  // cerr << "\n\nin exit_bad_status(): '" << raw_resp << "'\n\n";

  uint err_code{0};
  string err_msg{};
  vector<pair<string, uint>> err_codes{{"403 Forbidden", 403},
                                       {"404 Not Found", 404},
                                       {"500 Internal Server Error", 500},
                                       {"502 Bad Gateway", 502}};

  // Search for an error code inside the response
  const auto& low_resp{lower_cp(raw_resp)};
  for (const auto& [msg, code] : err_codes) {
    auto resp_it{low_resp.find(lower_cp(msg))};
    if (resp_it != string::npos) {
      err_msg = msg;
      err_code = code;
      break;
    }
  }

  alert_user(parent->page_uri, err_msg);
  if (err_code)
    printf("HTTP status code: %u\n", err_code);
  else
    printf("\n");
}

//------------------------------------------------------------------------------

// (Create || parse) the file 404's json data for this thread
void Page::load_404s()
{
  f404s_fpath = (targ_dir / "file_404s.json");
  if (!fs::exists(f404s_fpath))
    ofstream cfg_ofs{f404s_fpath.string()};
  f404s_json = rd_json_file(f404s_fpath);

  // only load up if we aren't forcing the file check regardless
  if (!force_chk)
    for (const auto& file : f404s_json["404s"])
      file_404s.push_back(file.asString());
  else
    f404s_json.clear();  // this will be overwritten by the newer 404s
}

//------------------------------------------------------------------------------

// Write the config's json back out into the 'file_404s.json' file
void Page::write_404s() const
{
  // DESIGN: Just use an ofstream here now?
  fstream f404s_ofs{f404s_fpath.string(), ios::out};  // DESIGN: | ios::trunc ?
  fast_wrt_strm(f404s_ofs, f404s_json);
}

//------------------------------------------------------------------------------

// Set the specific JSON tag names for this page
void Page::set_board_tags()
{
  // DESIGN: Convert to vector<pair<string,string>> storage instead.
  file_ext_tag = brd_json["file_ext_tag"].asString();
  file_name_tag = brd_json["file_name_tag"].asString();
  orig_file_tag = brd_json["file_orig_tag"].asString();
  files_arr_tag = brd_json["files_arr_tag"].asString();
  op_replies_tag = brd_json["op_replies_tag"].asString();
  board_type = brd_json["board_type"].asString();
}

//------------------------------------------------------------------------------

// Store all a thread's media files into it's local directory
void Page::grab_thrd_files()
{
  const fs::path thumbs_dir{targ_dir / ".thumbs"};
  fs::create_directory(thumbs_dir);

  const fs::path media_dir{targ_dir / "media"};
  fs::create_directory(media_dir);

  if (!json().empty()) {  // in case of parse error in load()
    //
    if ((board_type == "openib") || (board_type == "vichan")) {

      // DESIGN: code-envy w/ grab_files()?

      const auto posts{json()[op_replies_tag]};
      for (const auto& post : posts) {

        const auto orig{post[orig_file_tag].asString()};
        if (orig != "")
          grab_file(post, orig, media_dir);

        const auto files_arr{post[files_arr_tag]};
        if (files_arr != Json::nullValue) {

          for (const auto& file : files_arr) {
            const auto orig{file[orig_file_tag].asString()};
            if (orig != "")
              grab_file(file, orig, media_dir);
          }
        }
      }
    }
    //
    else {
      grab_files(json()[files_arr_tag], media_dir);

      const auto posts{json()[op_replies_tag]};
      for (const auto& post : posts) {
        grab_files(post[files_arr_tag], media_dir);
      }
    }
  }

  write_404s();
}

//------------------------------------------------------------------------------

// Save the thread's media file array files (as needed)
uint Page::grab_files(const Json::Value& files_arr, const fs::path& media_dir)
{
  uint file_cnt{0};

  for (const auto& file : files_arr) {
    const auto orig_file{file[orig_file_tag].asString()};
    grab_file(file, orig_file, media_dir);
  }

  return file_cnt;
}

//------------------------------------------------------------------------------

void Page::grab_file(const Json::Value& file,
                     const std::string& orig,
                     const fs::path& media_dir)
{
  // DESIGN: Replace '&apos;' w/ ' in filenames?
  //  -does the frequency of the issue warrant the on-going compute cost
  //   4ever?
  // orig = repl_w_char(orig, "&apos;", '\'');

  const auto ext{file[file_ext_tag].asString()};
  const auto orig_file{orig + ext};  //
  // DESIGN: rework this to not rely on a returned filepath, but merely a
  // (truncated) string instead. Assemble the path here. Other functions need
  // this truncate functionality w/o the filepath association.
  const fs::path media_file{trunc_filename(orig_file, media_dir)};

  bool attempt_dl{false};
  if (fs::exists(media_file)) {
    // DESIGN: Manage duplicate names; when not duplicate files (eg
    // 'ClipboardFile.png')

    // Beware. This will be the '98% zone' after an archive is first built.
    // -stay lightweight.
  }
  else
    attempt_dl = true;

  // Make sure the file isn't already in the 404 list
  const bool have_404{std::find(begin(file_404s), end(file_404s), orig_file) !=
                      end(file_404s)};

  if ((!have_404 && attempt_dl) || force_chk) {
    std::cout << "save: " << orig_file << "\n";

    auto path_of_file{file[file_name_tag].asString() + ext};

    // Adapt the file's URI request to each imageboard type
    string file_uri{};
    //
    if (board_type == "vichan")
      file_uri = muh_curl::muh_file_url(site_nm + "/" + board_nm, path_of_file);
    else if (board_type == "openib")
      file_uri = muh_curl::muh_file_url("media." + site_nm, path_of_file);
    else
      file_uri = muh_curl::muh_file_url(site_nm, path_of_file);

    auto file_resp = muh_curl::get_resp(file_uri);

    // Attempt to save the file_resp out as the original filename
    if (file_resp.rdbuf()->in_avail() != 0) {

      // DESIGN: refactor out this small block
      // Check size of stream in bytes (for 404's, etc)
      file_resp.seekg(0, ios::end);
      const auto sz_bytes{file_resp.tellg()};
      file_resp.seekg(0, ios::beg);
      //
      uint err_code{};
      if (sz_bytes < 200)  // response size is too small, check for error page
        err_code = chk_resp_err(file_resp.str());

      if (!err_code) {
        printf("         <==  %s\n", file_uri.c_str());

        ofstream ofsb(media_file, std::ios::binary);
        ofsb << file_resp.rdbuf();
      }
      else {
        std::cout << err_code << '\n';

        if (err_code == 404) {
          Json::Value elem{orig_file};
          f404s_json["404s"].append(elem);
        }
      }
    }
    else
      std::cout << "empty\n";
  }
}

//------------------------------------------------------------------------------

// DESIGN: This shouldn't be returning a path but simply a (truncated) string.

// Ensure the file name isn't too long
fs::path Page::trunc_filename(const std::string& orig_file,
                              const fs::path& media_dir) const
{
  fs::path res{};
  // DESIGN: include a trim() on res before return.

  if (orig_file.size() >= 120) {  // Limit the length of local filenames
    auto [filename, ext]{split_parts(orig_file)};
    filename = filename.substr(0, 115);
    // If we have a file extension from the split then use it
    res = (media_dir / (filename + (ext != "" ? "." + ext : "")));
  }
  else
    res = (media_dir / orig_file);

  return res;
}

//------------------------------------------------------------------------------

// First get the html response, then store both to disk
void Page::wrt_json_html()
{
  resp("html");

  write_file(page_nm, "json", targ_dir, raw_json);
  write_file(page_nm, "html", targ_dir, raw_html);
}

//------------------------------------------------------------------------------

// Overwrite various types of file data into a file
template<typename T>
void Page::write_file(const std::string& file_nm,
                      const std::string& ext,
                      const fs::path& targ_dir,
                      const T& file_dat) const
{
  const fs::path filename{file_nm + "." + ext};
  const fs::path filepath{targ_dir / filename};
  ofstream ofs(filepath.string());
  ofs << file_dat;
}

//------------------------------------------------------------------------------

Json::Value Page::json() const
{
  return jsondb.db;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

// Parameterized Bump ctor
Bump::Bump(const int argc, char** const argv)
{
  std::tie(site_nm, board_nm, force_chk) = parse_args(argc, argv);

  arch_dir = ".";  // DESIGN: This can be added into the args list in the future

  init_arch();
  set_board_tags();

  // Re-load the class' 'catalog' member field now that we have the needed
  // data properly available
  catalog = Page{
      site_nm, board_nm, "catalog", board_dir, file_frc_chk, board_json};
  if (catalog.json().empty())
    std::exit(255);  // w/ a bad catalog, there's no point in proceeding.

  chk_version();

  chk_load_bumps();
}

//------------------------------------------------------------------------------

// Parse the user-supplied arguments for the program
std::tuple<std::string, std::string, bool> Bump::parse_args(const int argc,
                                                            char** const argv)
{
  if (argc < 3) {
    cerr << "\nprogram argument error\n"
         << "usage: " << argv[0] << "  site_name  board_name\n";

    std::exit(-1);
  }

  const string site_nm{argv[1]};
  const string board_nm{argv[2]};

  // Optional 3rd flag argument to force full re-checking of the archive
  bool force_chk{false};
  if (argv[3]) {
    istringstream bool_arg{argv[3]};

    if (bool_arg.str() == "true")
      bool_arg >> std::boolalpha >> force_chk;
    else if (bool_arg.str() == "1")
      bool_arg >> force_chk;
    else {
      cerr << "\n'force check' flag argument error\n"
           << "usage: 'true' or '1'\n";

      std::exit(-1);
    }
  }

  return make_tuple(site_nm, board_nm, force_chk);
}

//------------------------------------------------------------------------------

// Initialize the archive in general
void Bump::init_arch()
{
  fs::path sites_fpath{arch_dir / ".sites.config"};
  sites_json = rd_json_file(sites_fpath);
  if (!cnfrm_good_site()) {
    cerr << "\nSite: '" << site_nm
         << "' is not basedt in the technical '.sites.config' list.  Is a new"
            "entry needed?\n\n";  // :^)
    std::exit(127);
  }

  std::tie(board_dir, thrds_dir) = init_arch_dirs();

  // (Create || parse) the archbot config's json data
  cfg_fpath = (board_dir / ".archbot.config");
  if (!fs::exists(cfg_fpath))
    ofstream cfg_ofs{cfg_fpath.string()};
  cfg_json = rd_json_file(cfg_fpath);
}

//------------------------------------------------------------------------------
// clang-format off
// Look for the site's name in the good 'all_json_sites' list  :^)
bool Bump::cnfrm_good_site()
{bool research_into_these_matters_of_importance_tbh{
false};bool                            dindu_nuffin{false};for(const
                                       auto&e:
sites_json["all_json_sites"])if(site_nm                                     ==
e["site_name"].asString(                             ))//      good site found
                                                                            {
  if (e["board_type"].asString() == "jschan")
muh_curl::res_dir=                  "thread";
                                     muh_curl::media_dir  =e["media_dir_tag"]  .
asString();
research_into_these_matters_of_importance_tbh=true;}for(const  auto&septic   :
sites_json["blacklist_sites"])                        if(      site_nm
==septic["site_name"].                          asString( ))// evil site found
research_into_these_matters_of_importance_tbh
=                                                      false
;else{                                                 Json::Value
purple_dranks{};Json::Value             rainbow_dash_is_clearly_best_pone_tbh{
};rainbow_dash_is_clearly_best_pone_tbh
["site_name"]="spacechan.xyz";Json::Value skittles{};skittles[     "site_name"]
="bunkerchan.xyz";Json::Value cant_corner_the_dorner{}; cant_corner_the_dorner[
 "site_name"                              ]="16chan.xyz";purple_dranks.append(
rainbow_dash_is_clearly_best_pone_tbh);purple_dranks.append(          skittles)
;purple_dranks                   .                   append(
cant_corner_the_dorner
                                 )                                     ;
for(const auto&he_a_gud_boy:purple_dranks)//             __really__good
//boysitefound_in_fact_of_course.                                   :^)
if(site_nm==                                  he_a_gud_boy["site_name"]
.asString                                           (             )            )
research_into_these_matters_of_importance_tbh=dindu_nuffin;}       return
research_into_these_matters_of_importance_tbh                        ;}
// clang-format on
//------------------------------------------------------------------------------

// Initialize the archive's basic directory structure within the filesystem
std::tuple<fs::path, fs::path> Bump::init_arch_dirs()
{
  const fs::path site_dir{site_nm};
  fs::create_directory(arch_dir / site_dir);
  const fs::path board_dir{site_dir / board_nm};
  fs::create_directory(board_dir);
  const fs::path thrds_dir{board_dir / "threads"};
  fs::create_directory(thrds_dir);
  const fs::path static_dir{board_dir / ".static"};
  fs::create_directory(static_dir);

  return make_tuple(board_dir, thrds_dir);
}

//------------------------------------------------------------------------------

// Set the specific JSON tag names for this board. This 'x_tag's approach
// allows BUMP to efficiently support differing IB software JSON schemas.
void Bump::set_board_tags()
{
  for (const auto& site : sites_json["all_json_sites"]) {

    if (site_nm == site["site_name"].asString()) {

      // DESIGN: Convert to vector<pair<string,string>> storage instead.

      // These are used inside Page
      board_json["file_ext_tag"] = site["file_ext_tag"].asString();
      board_json["file_name_tag"] = site["file_name_tag"].asString();
      board_json["file_orig_tag"] = site["file_orig_tag"].asString();
      board_json["files_arr_tag"] = site["files_arr_tag"].asString();
      board_json["op_replies_tag"] = site["op_replies_tag"].asString();
      board_json["board_type"] = site["board_type"].asString();

      // These are used inside Bump
      thread_id_tag = site["thread_id_tag"].asString();
      subject_tag = site["subject_tag"].asString();
      comment_tag = site["comment_tag"].asString();
      last_bump_tag = site["last_bump_tag"].asString();
      board_type = site["board_type"].asString();
      // board_json["media_dir_tag"] = site["media_dir_tag"].asString();

      break;
    }
  }
}

//------------------------------------------------------------------------------

// Check the version number of the config, rebuilding the file if necessary
void Bump::chk_version()
{
  const auto cfg_vrsn{cfg_json["version"].asString()};
  if (cfg_vrsn != bump_vrsn) {

    // DESIGN: pick out specific version diffs that should trigger a cfg
    // rebuild

    // <= bump_v0.1b; prior to versioning, so rebuild the config
    // <= bump_v0.2; prior to dir naming changes, so rebuild the config
    if (cfg_vrsn <= "bump_v0.2")
      cfg_rebuild();

    cfg_json["version"] = bump_vrsn;
  }
}

//------------------------------------------------------------------------------

// This will force a rebuild of the .archbot.config file
void Bump::cfg_rebuild()
{
  ofstream cfg_ofs{cfg_fpath.string()};
  cfg_json.clear();
  force_chk = true;
}

//------------------------------------------------------------------------------

// Check if we have any bumps, then compare live and archive threads together
void Bump::chk_load_bumps()
{
  // // First check if we have any bumps at all
  load_live_thrds();
  cmp_last_bumps();

  // If so, proceed with pre-processing
  if (force_chk || have_bumps) {
    catalog.wrt_json_html();
    find_all_bumps();
  }
}

//------------------------------------------------------------------------------

// Load the live_thrds vector from the catalog json
void Bump::load_live_thrds()
{
  if ((board_type == "vichan") || (board_type == "openib"))
    // spin the outer page array
    for (const auto& page : catalog.json())
      live_loop(page["threads"]);

  else
    live_loop(catalog.json());
}

//------------------------------------------------------------------------------

void Bump::live_loop(const Json::Value& thrds_arry)
{
  for (const auto& thrd : thrds_arry)
    live_thrds.emplace_back(make_tuple(thrd[thread_id_tag].asUInt(),
                                       thrd[subject_tag].asString(),
                                       thrd[last_bump_tag].asString(),
                                       thrd[comment_tag].asString()));
}

//------------------------------------------------------------------------------

// Compare the live board's 'last known bump' vs. the archive's
void Bump::cmp_last_bumps()
{
  // First thing, sort out the 'lastBump's from the board
  const auto live_bumps_srtd{sort_live_bumps()};
  const auto live_last_bump{live_bumps_srtd[0]};  // latest bump will be 0-th

  const auto cfg_last_bump{cfg_json["last_known_bump"].asString()};

  if (live_last_bump != cfg_last_bump) {  // We have a new last known bump
    have_bumps = true;
    cfg_json["last_known_bump"] = live_last_bump;
  }
  else
    have_bumps = false;
}

//------------------------------------------------------------------------------

// Sort the 'lastBump's from the live_thrds, descending. (newest first)
std::vector<std::string> Bump::sort_live_bumps()
{
  vector<string> live_bumps{};

  for (const auto& thrd : live_thrds)
    live_bumps.emplace_back(get<2>(thrd));  // bump field is 2-th tuple element

  sort(begin(live_bumps), end(live_bumps), std::greater<string>());

  return live_bumps;
}

//------------------------------------------------------------------------------

// Collect up all the newly-bumped thread IDs
void Bump::find_all_bumps()
{
  sort_live_thrds();
  load_cfg_thrds();

  cmp_all_bumps();
}

//------------------------------------------------------------------------------

// Sort the live_thrds vector by IDs
void Bump::sort_live_thrds()
{
  sort(begin(live_thrds),
       end(live_thrds),
       [](const auto& thrda, const auto& thrdb) -> bool {
         return get<0>(thrda) < get<0>(thrdb);  // 'id' is 0-th tuple element
       });
}

//------------------------------------------------------------------------------

// Load the pre-sorted cfg_thrds vector
void Bump::load_cfg_thrds()
{
  // The "threads" json array loaded from the .archbot.config is already
  // sorted by ID, so simply load up the vector.
  for (const auto& thrd : cfg_json["threads"])
    cfg_thrds.emplace_back(make_tuple(thrd[thread_id_tag].asUInt(),
                                      thrd[subject_tag].asString(),
                                      thrd[last_bump_tag].asString(),
                                      thrd["subj_id_dir"].asString()));
}

//------------------------------------------------------------------------------

// Compare all the bumps from the live board and the archive, storing the
// difference between them (ie, keep the newly bumped IDs)
void Bump::cmp_all_bumps()
{
  // DESIGN: What about the case for thread deletions?
  //  - Dealing with deletions may require confirmation of IDs as well.

  // Ensure the cfg_thrds size is identical, in case live_thrds has new
  // threads.
  const auto live_sz{live_thrds.size()};
  const auto cfg_sz{cfg_thrds.size()};
  const bool have_diff{live_sz != cfg_sz};
  if (have_diff)
    cfg_thrds.resize(live_sz);

  // Compare the live and config 'lastBump's side-by-side.
  for (size_t i{0}; i < live_sz; ++i) {
    const auto& [live_id, subj, live_bump, cmnt]{live_thrds[i]};
    const auto& cfg_bump{get<2>(cfg_thrds[i])};  // 'bump' is 2-th tuple element

    if (force_chk || (live_bump != cfg_bump))
      bump_ids.emplace_back(live_id);
  }

  if (have_diff)               // Now, remove any extraneous elements that were
    cfg_thrds.resize(cfg_sz);  // added just to prepare for the comparision loop
}

//------------------------------------------------------------------------------

std::string Bump::site() const
{
  return site_nm;
}

std::string Bump::name() const
{
  return board_nm;
}

bool Bump::has_bumps() const
{
  return have_bumps;
}

std::vector<uint> Bump::bumps() const
{
  return bump_ids;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

// Return the thread vector's iterator if the ID is found inside the element
template<typename T> auto Bump::locate_id(const uint id, T& thrd_vec)
{
  return find_if(
      begin(thrd_vec), end(thrd_vec), [id](const auto& thrd) -> bool {
        // 'id' is in the 0-th position of the tuple
        return get<0>(thrd) == id;
      });
}

//------------------------------------------------------------------------------

// For any threads that are bumped, save all the files out (as needed),
// creating custom-named directories for any new threads.
void Bump::save_bump_thrds()
{
  if (force_chk)
    cerr << "\nChecking all threads...\n\n";

  // Iteratively check if each bumped thread already has it's special
  // directory yet, and create one if not (that is, its a new thread and needs
  // a folder).
  for (const auto bump_id : bump_ids) {
    fs::path dir{};

    auto cfg_it = locate_id(bump_id, cfg_thrds);
    if (!force_chk &&
        (cfg_it != end(cfg_thrds))) {  // Thread's directory already created
      auto& [id, subj, cfg_bump, cfg_dir]{*cfg_it};
      dir = cfg_dir;

      // Get the live thread's bump and store it back into the config data
      const auto live_it = locate_id(bump_id, live_thrds);
      const auto& live_bump{get<2>(*live_it)};  // 'bump' is 2-th tuple element
      cfg_bump = live_bump;
    }
    //
    else {  // ID not found (ie, have a new thread) so need to create new dir

      // Get the new thread's data and create a custom dir from it's subj & id
      const auto live_it = locate_id(bump_id, live_thrds);
      const auto& [id, subj, bump, cmnt]{*live_it};
      //
      if (subj != "")
        dir = add_thrd_dir(subj, bump_id);
      //
      else {  // use OP comment instead
        std::string cmnt_cp{cmnt};
        if (cmnt_cp.size() > 200)  // no need to send an overly long comment
          cmnt_cp = cmnt_cp.substr(0, 200);

        dir = add_thrd_dir(cmnt_cp, bump_id);
      }

      // Retain for updating back into the .archbot.config, including the new
      // dir path
      cfg_thrds.emplace_back(make_tuple(bump_id, subj, bump, dir.string()));
    }

    // Save the thread's files
    Page thrd{
        site_nm, board_nm, to_string(bump_id), dir, file_frc_chk, board_json};
    thrd.wrt_json_html();
    thrd.grab_thrd_files();
  }

  update_cfg();
}

//------------------------------------------------------------------------------

// Store the updated thread data back into the config's json.
void Bump::update_cfg()
{
  cfg_json["threads"].clear();
  for (const auto& [id, subj, bump, dir] : cfg_thrds) {
    Json::Value elem{};
    elem[thread_id_tag] = id;
    elem[subject_tag] = subj;
    elem[last_bump_tag] = bump;
    elem["subj_id_dir"] = dir;

    cfg_json["threads"].append(elem);
  }

  write_cfg();
}

//------------------------------------------------------------------------------

// Write the config's json back out into the .archbot.config file
void Bump::write_cfg()
{
  // DESIGN: Just use an ofstream here now?
  fstream cfg_ofs{cfg_fpath.string(), ios::out};  // DESIGN: | ios::trunc ?
  fast_wrt_strm(cfg_ofs, cfg_json);
}

//------------------------------------------------------------------------------

// Given a thread's subject & id, create a custom directory for it under the
// general threads directory
fs::path Bump::add_thrd_dir(const std::string& subj, const uint id)
{
  const auto clean_subj{make_dir_name(subj)};
  const auto id_pad{make_id_pad(id)};
  const fs::path subj_pad{clean_subj + "_" + id_pad};
  const fs::path thrd_dir{thrds_dir / subj_pad};
  fs::create_directory(thrd_dir);

  return thrd_dir;
}

//------------------------------------------------------------------------------

// Clean up the thread's subj/cmnt suitable for use as a portable directory
// name
std::string Bump::make_dir_name(const std::string& subj)
{
  string res{subj};

  const vector<string> bad_strings{"&quot;",
                                   "&amp;",
                                   "&apos;",
                                   "&gt;",
                                   "<p class=\"body-line empty \">",
                                   "<p class=\"body-line ltr \">",
                                   "<p class=\"body-line ltr quote\">",
                                   "<span class=\"heading\">",
                                   "<span class=\"spoiler\">",
                                   "<span class=\"quote\">",
                                   "<a href=",  // TODO: master regex :^)
                                   "https",     // ..
                                   "http",      // ..
                                   "<a onclick=\"highlightReply('",  // ..
                                   "', event);\" href=\"/",          // ..
                                   "/res/",                          // ..
                                   "<em>",
                                   "</em>",
                                   "<br/>",
                                   "</span>"};
  for (const auto& bad_string : bad_strings)
    zap_bad_string(res, bad_string);

  const vector<string> spc_strings{"\r\n", "</p>"};
  for (const auto& spc_string : spc_strings)
    res = repl_w_char(res, spc_string, ' ');

  rem_dups(res, ' ');

  const vector<pair<char, char>> bad_goods{
      {' ', '_'},  // intentionally first
      {'/', '_'},  {'.', '_'}, {'!', '_'},  {'?', '_'},  {'&', 'n'}, {'^', ' '},
      {'$', ' '},  {'%', ' '}, {'#', ' '},  {'*', ' '},  {'~', ' '}, {'"', ' '},
      {'\'', ' '}, {'=', ' '}, {'\0', ' '}, {'\n', ' '}, {'(', ' '}, {')', ' '},
      {'<', ' '},  {'>', ' '}, {',', ' '},  {':', ' '},  {';', ' '}};

  for (const auto& bad_good : bad_goods)
    repl_bad_char(res, bad_good);

  zap_bad_string(res, " ");

  rem_dups(res, '_');

  res = trim_char(res, '_');  // rem leading or trailing '_'

  // Limit the length of local directory names (47 + '_1234567' == 55 total)
  if (res.size() > 47)
    res = res.substr(0, 47);

  res = trim_char(res, '_');

  if ((res.size() == 0) || res == "\u200c")  // half-space le maymay subj.
    res = "no_u";                            // :^)

  return res;
}

//------------------------------------------------------------------------------

// Remove bad sub-strings
void Bump::zap_bad_string(std::string& in_out, const std::string& bad)
{
  using str_sz_t = string::size_type;
  str_sz_t bad_len = bad.length();
  const auto str_end{string::npos};

  for (str_sz_t i = in_out.find(bad); i != str_end; i = in_out.find(bad)) {
    in_out.erase(i, bad_len);
  }
}

//------------------------------------------------------------------------------

// Replace bad chars with a good char
void Bump::repl_bad_char(std::string& in_out,
                         const std::pair<char, char> bad_good)
{
  const auto [bad, good]{bad_good};

  std::transform(
      begin(in_out),
      end(in_out),
      begin(in_out),
      [=](unsigned char c) -> unsigned char { return c == bad ? good : c; });
}

//------------------------------------------------------------------------------

// Pad the ID out to n digits, using zeros
std::string Bump::make_id_pad(const uint id, const uint n)
{
  std::stringstream res{};
  res << std::setfill('0') << std::setw(n) << id;
  return res.str();
}

// Copyright (2019)
// License (MIT) https://opensource.org/licenses/MIT
