// BUMP, the imageboard porter
// ===========================
// -This software is designed to help out anon with imageboards

// -Filename: Bump.h
// -The declarations of the Bump class, and it's ancillary Page class

#pragma once

#include <map>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

// The older compiler has filesystem under the experimental namespace
// #if __GNUC__ == 7
// #  include <experimental/filesystem>
// namespace fs = std::experimental::filesystem;
// #elif __GNUC__ >= 8
#include <filesystem>
namespace fs = std::filesystem;
// #endif

#include <jsoncpp/json/json.h>
// #include <json/json.h>  // <= you might need this location instead

//------------------------------------------------------------------------------

// Class for storing individual page data; it's a field inside the Bump class
class Page {
 public:
  Page();
  Page(const std::string& site_name,
       const std::string& board_name,
       const std::string& page_name,
       const fs::path& target_dir,
       bool force_check,
       Json::Value board_json);

  Json::Value json() const;

  void grab_thrd_files();

  void wrt_json_html();

 private:
  void load_json();

  void load_404s();

  void resp(const std::string& page_tp);

  void set_board_tags();

  uint grab_files(const Json::Value& files_arr, const fs::path& media_dir);

  void grab_file(const Json::Value& post,
                 const std::string& orig,
                 const fs::path& media_dir);

  fs::path trunc_filename(const std::string& orig_file,
                          const fs::path& media_dir) const;

  // std::tuple<std::string, std::string> split_file_ext(
  //     const std::string& filename) const;

  template<typename T>
  void write_file(const std::string& file_nm,
                  const std::string& ext,
                  const fs::path& targ_dir,
                  const T& file_dat) const;

  std::string site_nm;
  std::string board_nm;
  std::string page_nm;
  fs::path targ_dir;
  bool force_chk;

  std::string page_uri;
  std::string raw_json;
  std::string raw_html;

  Json::Value f404s_json;
  std::vector<std::string> file_404s;
  fs::path f404s_fpath;
  void write_404s() const;

  // DESIGN: Convert to vector<pair<string,string>> storage instead.
  Json::Value brd_json;
  //
  std::string file_ext_tag;
  std::string file_name_tag;
  std::string orig_file_tag;
  std::string files_arr_tag;
  std::string op_replies_tag;
  std::string board_type;

  // Internal for storing parsed json
  struct Jsondb {
    void load(const std::string& raw_json);
    Json::Value db;

    const Page* parent;
    void rpt_bad_status(const std::string& raw_json) const;
  } jsondb;
};

//------------------------------------------------------------------------------

// The primary BUMP class
class Bump {
  const std::string bump_vrsn{
      "bump_v0.2"};  // Need to set tag bool-op in chk_version()?

 public:
  Bump(const int argc, char** const argv);

  std::string site() const;

  std::string name() const;

  bool has_bumps() const;

  std::vector<uint> bumps() const;

  void save_bump_thrds();

 private:
  //--- pre save_bump_thrds() ---//

  std::tuple<std::string, std::string, bool> parse_args(const int argc,
                                                        char** const argv);
  void init_arch();

  bool cnfrm_good_site();

  std::tuple<fs::path, fs::path> init_arch_dirs();

  void set_board_tags();

  void chk_version();

  void cfg_rebuild();

  void chk_load_bumps();

  void load_live_thrds();

  void live_loop(const Json::Value& thrds_arry);

  void sort_live_thrds();

  void load_cfg_thrds();

  void find_all_bumps();

  void cmp_last_bumps();

  std::vector<std::string> sort_live_bumps();

  void cmp_all_bumps();

  std::string site_nm;
  std::string board_nm;
  bool force_chk;
  bool file_frc_chk;

  fs::path arch_dir;
  fs::path board_dir;
  fs::path thrds_dir;

  Json::Value sites_json;

  // DESIGN: Convert to vector<pair<string,string>> storage instead.
  Json::Value board_json;
  std::string thread_id_tag;
  std::string subject_tag;
  std::string comment_tag;
  std::string last_bump_tag;
  std::string board_type;

  fs::path cfg_fpath;
  Json::Value cfg_json;

  Page catalog;

  bool have_bumps;
  std::vector<uint> bump_ids;

  const std::vector<std::map<std::string,
                             std::tuple<std::pair<std::string, std::string>,
                                        std::pair<std::string, std::string>,
                                        std::pair<std::string, std::string>>>>
      site_json_maps;

  std::vector<std::tuple<uint, std::string, std::string, std::string>>
      live_thrds;
  std::vector<std::tuple<uint, std::string, std::string, std::string>>
      cfg_thrds;

  //--- post save_bump_thrds() ---//

  template<typename T> auto locate_id(const uint id, T& thrd_vec);

  void update_cfg();

  void write_cfg();

  fs::path add_thrd_dir(const std::string& subj, const uint id);

  std::string make_dir_name(const std::string& subj);

  void zap_bad_string(std::string& in_out, const std::string& bad);

  void repl_bad_char(std::string& in_out, const std::pair<char, char> bad_good);

  std::string make_id_pad(const uint id, const uint n = 7);

  std::vector<Page> bump_thrds;
};

// Copyright (2019)
// License (MIT) https://opensource.org/licenses/MIT
